@@ -9,15 +9,21 @@
 import abc
 import functools
 import json
+import os
 import socket
 import time
+import traceback
+import warnings
 from contextlib import closing
+from dataclasses import dataclass, field
 from enum import Enum
-from typing import Any, Callable, Dict, List, Tuple
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union
 
 import torchelastic.rendezvous as rdzv
 import torchelastic.utils.store as store_util
+from torchelastic.events import record, Event, EventSource
 from torchelastic.metrics import prof, put_metric
+from torchelastic.multiprocessing import ProcessFailure, Std
 from torchelastic.utils.logging import get_logger
 
 
@@ -27,64 +33,69 @@ DEFAULT_ROLE = "default"
 log = get_logger()
 
 
+@dataclass
 class WorkerSpec:
     """
     Contains blueprint information about a particular type of worker.
     For a given role, there must only exist a single worker spec.
     Worker spec is expected to be homogenous across all nodes (machine),
     that is each node runs the same number of workers for a particular spec.
+
+    Args:
+        role: user-defined role for the workers with this spec
+        local_world_size: number local workers to run
+        fn: (deprecated use entrypoint instead)
+        entrypoint: worker function or command
+        args: arguments to pass to ``entrypoint``
+        rdzv_handler: handles rdzv for this set of workers
+        max_restarts: number of max retries for the workers
+        monitor_interval: monitor status of workers every ``n`` seconds
+        master_port: fixed port to run the c10d store on rank 0
+                     if not specified then will chose a random free port
+        redirects: redirect std streams to a file,
+                   selectively redirect for a particular
+                   local rank by passing a map
+        tee: tees the specified std stream(s) to console + file,
+             selectively tee for a particular local rank by passing a map,
+             takes precedence over ``redirects`` settings.
+
     """
 
-    __slots__ = [
-        "role",
-        "local_world_size",
-        "fn",
-        "args",
-        "rdzv_handler",
-        "max_restarts",
-        "monitor_interval",
-        "master_port",
-    ]
+    role: str
+    local_world_size: int
+    rdzv_handler: rdzv.RendezvousHandler
+    fn: Optional[Callable] = None
+    # TODO @kiuk - make entrypoint a required field
+    entrypoint: Union[Callable, str, None] = None
+    args: Tuple = ()
+    max_restarts: int = 3
+    monitor_interval: float = 30.0
+    master_port: Optional[int] = None
+    redirects: Union[Std, Dict[int, Std]] = Std.NONE
+    tee: Union[Std, Dict[int, Std]] = Std.NONE
 
-    def __init__(
-        self,
-        role: str,
-        local_world_size: int,
-        fn: Callable,
-        args: Tuple,
-        rdzv_handler: rdzv.RendezvousHandler,
-        max_restarts: int = 3,
-        monitor_interval: float = 30.0,
-        master_port=None,
-    ):
-        r"""
+    def __post_init__(self):
+        assert self.local_world_size > 0
+        assert self.monitor_interval > 0
 
-        Arguments:
-            role (str): user-defined role for the workers with this spec
-            local_world_size (int): number local workers to run
-            fn (Callable): worker main entry point function
-            args (Tuple): arguments to pass to ``fn(args)``
-            rdzv_handler (RendezvousHandler): handles rdzv for this set of workers
-            max_restarts (int): number of max retries for the workers
-            monitor_interval (int): monitor status of workers every ``n`` seconds
-            master_port (int): fixed port to run the c10d store on rank 0
-                               if not specified then will chose a random free port
+        if self.fn:
+            warnings.warn(
+                "WorkerSpec.fn will be deprecated,"
+                " please use WorkerSpec.entrypoint instead",
+                category=DeprecationWarning,
+            )
+            self.entrypoint = self.fn
+        assert self.entrypoint
+
+    def get_entrypoint_name(self):
         """
-
-        assert local_world_size > 0
-        assert max_restarts >= 0
-        assert monitor_interval > 0
-
-        # Note: role is not used for data parallel, every worker has the same role
-        # wiring it in to handle more elaborate situations later
-        self.role = role
-        self.local_world_size = local_world_size
-        self.fn = fn
-        self.args = args
-        self.rdzv_handler = rdzv_handler
-        self.max_restarts = max_restarts
-        self.monitor_interval = monitor_interval
-        self.master_port = master_port
+        If the entrypoint is a function (e.g. ``Callable``) returns its ``__qualname__``,
+        else if the entrypoint is a binary (e.g. ``str``), returns the binary name.
+        """
+        if isinstance(self.entrypoint, str):
+            return os.path.basename(self.entrypoint)
+        else:
+            return self.entrypoint.__qualname__
 
 
 class Worker:
@@ -99,7 +110,7 @@ class Worker:
     agent, it could be the ``pid (int)`` of the worker, for a remote
     agent it could be encoded as ``host:port (string)``.
 
-    Arguments:
+    Args:
         id (Any): uniquely identifies a worker (interpreted by the agent)
         local_rank (int): local rank of the worker
         global_rank (int): global rank of the worker
@@ -161,7 +172,7 @@ class Worker:
         return str(self)
 
 
-class WorkerState(Enum):
+class WorkerState(str, Enum):
     """
     State of the ``WorkerGroup``. Workers in a worker group change state as a unit.
     If a single worker in a worker group fails the entire set is considered
@@ -196,20 +207,20 @@ class WorkerState(Enum):
     self terminating and allowing the job manager to retry the node.
     """
 
-    UNKNOWN = 0
-    INIT = 1
-    HEALTHY = 2
-    UNHEALTHY = 4
-    STOPPED = 8
-    SUCCEEDED = 16
-    FAILED = 32
+    UNKNOWN = "UNKNOWN"
+    INIT = "INIT"
+    HEALTHY = "HEALTHY"
+    UNHEALTHY = "UNHEALTHY"
+    STOPPED = "STOPPED"
+    SUCCEEDED = "SUCCEEDED"
+    FAILED = "FAILED"
 
     @staticmethod
     def is_running(state: "WorkerState") -> bool:
         """
         Returns:
-             `` True`` if the worker state represents workers still running
-              (e.g. that the process exists but not necessarily healthy).
+             True if the worker state represents workers still running
+             (e.g. that the process exists but not necessarily healthy).
         """
         return state in {WorkerState.HEALTHY, WorkerState.UNHEALTHY}
 
@@ -248,7 +259,7 @@ class _RoleInstanceInfo:
     def __init__(self, role: str, rank: int, local_world_size: int):
         r"""
 
-        Arguments:
+        Args:
             role (str): user-defined role for the workers with this spec
             rank (int): the rank of the agent
             local_world_size (int): number of local workers to run
@@ -293,52 +304,38 @@ class _RoleInstanceInfo:
         return (start_idx, end_idx)
 
 
-class MonitorResult:
+@dataclass
+class RunResult:
     """
-    Returned by the agent's ``_monitor_workers`` API. A holder object
-    that holds information about the monitoring results.
-    The ``ret_vals`` and ``exceptions`` field map each worker's
-    return value (output) and exceptions (if any) accordingly by
-    the workers global rank.
+    Results returned by the worker executions. Run results follow an "all-or-nothing" policy
+    where the run is successful if and only if ALL local workers managed by this agent
+    complete successfully.
 
-    ``state = SUCCEEDED`` will have ``ret_val``.
-    ``state = FAILED`` will have ``exceptions``.
-    For other states both these fields will be empty.
+    If the result is successful (e.g. ``is_failed() = False``) then the ``return_values``
+    field contains the outputs (return values) of the workers managed by THIS agent mapped
+    by their GLOBAL ranks. That is ``result.return_values[0]`` is the return value of
+    global rank 0.
+
+    .. note:: ``return_values`` are only meaningful for when the worker entrypoint
+              is a function. Workers specified as a binary entrypoint do not canonically
+              have a return value and the ``return_values`` field is meaningless and
+              may be empty.
+
+    If ``is_failed()`` returns ``True`` then the ``failures`` field contains the
+    failure information, again, mapped by the GLOBAL rank of the worker that failed.
+
+    The keys in ``return_values`` and ``failures`` are mutually exclusive, that is,
+    a worker's final state can only be one of: succeeded, failed. Workers intentionally
+    terminated by the agent according to the agent's restart policy, are not represented
+    in either ``return_values`` nor ``failures``.
     """
 
-    __slots__ = ["state", "ret_vals", "exceptions"]
+    state: WorkerState
+    return_values: Dict[int, Any] = field(default_factory=dict)
+    failures: Dict[int, ProcessFailure] = field(default_factory=dict)
 
-    def __init__(
-        self,
-        state: WorkerState,
-        # pyre-fixme[9]: ret_vals has type `Dict[int, typing.Any]`; used as `None`.
-        ret_vals: Dict[int, Any] = None,
-        # pyre-fixme[9]: exceptions has type `Dict[int, Exception]`; used as `None`.
-        exceptions: Dict[int, Exception] = None,
-    ):
-        self.state = state
-        self.ret_vals = ret_vals
-        self.exceptions = exceptions
-
-
-class WorkerGroupFailureException(Exception):
-    """
-    Thrown when the agent cannot or has given up trying to run the workers.
-    This is typically thrown:
-
-    1. Exceeded ``max_restarts``.
-    2. Workers fail with errors that are deemed ``NonRestartable``
-
-    When constructing this exception the underlying worker exceptions
-    are provided as a map of the worker's global rank to the exception.
-    """
-
-    def __init__(self, msg: str, worker_excs: Dict[int, Exception]):
-        super().__init__(msg)
-        self._worker_excs = worker_excs
-
-    def get_worker_exceptions(self) -> Dict[int, Exception]:
-        return self._worker_excs
+    def is_failed(self) -> bool:
+        return self.state == WorkerState.FAILED
 
 
 def _get_socket_with_port() -> socket.socket:
@@ -403,29 +400,27 @@ class ElasticAgent(abc.ABC):
     Usage
     ::
 
-     try:
-         results = agent.run()
-         return results[0] # return rank 0's results
-     except WorkerGroupFailureException as e:
-         exceptions = e.get_worker_exceptions()
-         log.exception(f"worker 0 failed with: {exceptions[0]}")
-     except Exception as e:
-         log.exception(f"error while running agent")
+     group_result = agent.run()
+      if group_result.is_failed():
+        # workers failed
+        failure = group_result.failures[0]
+        log.exception(f"worker 0 failed with exit code : {failure.exit_code}")
+      else:
+        return group_result.return_values[0] # return rank 0's results
 
     """
 
     @abc.abstractmethod
-    def run(self, role: str = DEFAULT_ROLE) -> Dict[int, Any]:
+    def run(self, role: str = DEFAULT_ROLE) -> RunResult:
         """
         Runs the agent, retrying the worker group on failures up to
         ``max_restarts``.
 
         Returns:
-            The return values for each worker mapped by the worker's global rank.
-            Empty if workers have void signature.
+            The result of the execution, containing the return values or
+            failure details for each worker mapped by the worker's global rank.
 
         Raises:
-            WorkerGroupFailureException - workers did not successfully run
             Exception - any other failures NOT related to worker process
         """
         raise NotImplementedError()
@@ -454,12 +449,9 @@ class SimpleElasticAgent(ElasticAgent):
         self._remaining_restarts = self._worker_group.spec.max_restarts
         self._store = None
         self._exit_barrier_timeout = exit_barrier_timeout
+        self._total_execution_time = 0
 
-    # pyre-fixme[14]: `get_worker_group` overrides method defined in `ElasticAgent`
-    #  inconsistently.
-    def get_worker_group(self) -> WorkerGroup:
-        # TODO return an RO copy (need to create an ROWorkerGroup and ROWorkerSpec
-        # since both these classes contain non-pure-data pointers - e.g. rdzv_handler)
+    def get_worker_group(self, role: str = DEFAULT_ROLE) -> WorkerGroup:
         return self._worker_group
 
     @abc.abstractmethod
@@ -483,13 +475,20 @@ class SimpleElasticAgent(ElasticAgent):
         raise NotImplementedError()
 
     @abc.abstractmethod
-    def _monitor_workers(self, worker_group: WorkerGroup) -> MonitorResult:
+    def _monitor_workers(self, worker_group: WorkerGroup) -> RunResult:
         r"""
         Checks on the workers for the ``worker_group`` and returns
         the new state of the worker group.
         """
         raise NotImplementedError()
 
+    @abc.abstractmethod
+    def _shutdown(self) -> None:
+        """
+        Cleans up any resources that were allocated during the agent's work.
+        """
+        raise NotImplementedError()
+
     @staticmethod
     def _set_master_addr_port(store, master_port):
         if master_port is None:
@@ -529,22 +528,19 @@ class SimpleElasticAgent(ElasticAgent):
             self._set_master_addr_port(store, spec.master_port)
         master_addr, master_port = self._get_master_addr_port(store)
         restart_count = spec.max_restarts - self._remaining_restarts
-        workers_info = {
-            "local_ranks": [worker.local_rank for worker in workers],
-            "global_ranks": [worker.global_rank for worker in workers],
-            "role_ranks": [worker.role_rank for worker in workers],
-            "world_size": workers[0].world_size,
-            "role_world_size": workers[0].role_world_size,
-        }
+
         log.info(
-            f"[{spec.role}] Rendezvous complete for workers.\n"
-            f"Result:\n"
-            f"\trestart_count={restart_count}\n"
-            f"\tgroup_rank={group_rank}\n"
-            f"\tgroup_world_size={group_world_size}\n"
-            f"\tmaster_addr={master_addr}\n"
-            f"\tmaster_port={master_port}\n"
-            f"\tworkers={workers_info}\n"
+            f"[{spec.role}] Rendezvous complete for workers. Result:\n"
+            f"  restart_count={restart_count}\n"
+            f"  master_addr={master_addr}\n"
+            f"  master_port={master_port}\n"
+            f"  group_rank={group_rank}\n"
+            f"  group_world_size={group_world_size}\n"
+            f"  local_ranks={[worker.local_rank for worker in workers]}\n"
+            f"  role_ranks={[worker.role_rank for worker in workers]}\n"
+            f"  global_ranks={[worker.global_rank for worker in workers]}\n"
+            f"  role_world_sizes={[worker.role_world_size for worker in workers]}\n"
+            f"  global_world_sizes={[worker.world_size for worker in workers]}\n"
         )
 
     def _get_ranks(
@@ -659,9 +655,9 @@ class SimpleElasticAgent(ElasticAgent):
 
         log.info(f"[{role}] Starting worker group")
         worker_ids = self._start_workers(worker_group)
-        for local_rank, id in worker_ids.items():
+        for local_rank, w_id in worker_ids.items():
             worker = worker_group.workers[local_rank]
-            worker.id = id
+            worker.id = w_id
 
         worker_group.state = WorkerState.HEALTHY
 
@@ -678,17 +674,85 @@ class SimpleElasticAgent(ElasticAgent):
         self._initialize_workers(worker_group)
 
     @prof
-    def run(self, role: str = DEFAULT_ROLE) -> Dict[int, Any]:
-        is_failed = False
+    def run(self, role: str = DEFAULT_ROLE) -> RunResult:
+        start_time = time.monotonic()
         try:
-            return self._invoke_run(role)
-        except Exception as e:
-            is_failed = True
-            raise
+            result = self._invoke_run(role)
+            self._total_execution_time = int(time.monotonic() - start_time)
+            self._record_metrics(result)
+            self._record_worker_events(result)
+            return result
         finally:
-            self._record_metrics(is_failed)
+            # record the execution time in case there were any exceptions during run.
+            self._total_execution_time = int(time.monotonic() - start_time)
+            self._shutdown()
 
-    def _record_metrics(self, is_failed: bool = False):
+    def get_agent_status_event(self, state: WorkerState) -> Event:
+        raw_error = traceback.format_exc() if state == WorkerState.FAILED else None
+        return self._construct_event(
+            state.value, EventSource.AGENT, raw_error=raw_error
+        )
+
+    def _record_worker_events(self, result: RunResult) -> None:
+        for worker in self._worker_group.workers:
+            failure = result.failures.get(worker.global_rank)
+            state: str = self._get_worker_state(worker, result)
+            raw_error = json.dumps(failure.error_file_data) if failure else None
+            record(self._construct_event(state, EventSource.WORKER, worker, raw_error))
+
+    def _get_worker_state(self, worker: Worker, result: RunResult) -> str:
+        failure = result.failures.get(worker.global_rank)
+        if result.state in {WorkerState.UNHEALTHY, WorkerState.FAILED} and not failure:
+            # The worker got terminated by the torchelastic agent via SIGTERM signal
+            return "TERMINATED"
+        elif failure or worker.global_rank in result.return_values:
+            return result.state.value
+        else:
+            raise ValueError(f"Unknow worker: {worker.global_rank}")
+
+    def _construct_event(
+        self,
+        state: str,
+        source: EventSource,
+        worker: Optional[Worker] = None,
+        raw_error: Optional[str] = None,
+    ) -> Event:
+        wg = self._worker_group
+        spec = wg.spec
+        md = {
+            "group_world_size": wg.group_world_size,
+            "entry_point": spec.get_entrypoint_name(),
+        }
+        if worker:
+            md["local_rank"] = (worker.local_rank,)
+            md["role_rank"] = (worker.role_rank,)
+            md["role_world_size"] = (worker.role_world_size,)
+            global_rank = worker.global_rank
+            worker_id = str(worker.id)
+        else:
+            global_rank = None
+            worker_id = None
+        md_str = json.dumps(md)
+        metadata = {
+            "run_id": spec.rdzv_handler.get_run_id(),
+            "global_rank": global_rank,
+            "group_rank": wg.group_rank,
+            "worker_id": worker_id,
+            "role": spec.role,
+            "hostname": _get_fq_hostname(),
+            "state": state,
+            "total_run_time": self._total_execution_time,
+            "rdzv_backend": spec.rdzv_handler.get_backend(),
+            "raw_error": raw_error,
+            "metadata": md_str,
+            "agent_restarts": spec.max_restarts - self._remaining_restarts,
+        }
+        return Event(
+            f"torchelastic.worker.status.{state}", source=source, metadata=metadata
+        )
+
+    def _record_metrics(self, group_results: RunResult):
+        is_failed = group_results.is_failed()
         self._record_flakiness_metric(is_failed)
         spec = self._worker_group.spec
         restarts_happened = self._remaining_restarts != spec.max_restarts
@@ -715,9 +779,6 @@ class SimpleElasticAgent(ElasticAgent):
 
     def _record_flakiness_metric(self, is_failed: bool = False):
         if is_failed:
-            if not isinstance(is_failed, WorkerGroupFailureException):
-                # Only user code can contribute into flakiness score
-                return
             flakiness = 100.0
         else:
             spec = self._worker_group.spec
@@ -728,13 +789,15 @@ class SimpleElasticAgent(ElasticAgent):
 
         put_metric(f"workers.{spec.role}.flakiness", int(flakiness))
 
-    def _invoke_run(self, role: str = DEFAULT_ROLE) -> Dict[int, Any]:
+    def _invoke_run(self, role: str = DEFAULT_ROLE) -> RunResult:
         # NOTE: currently only works for a single role
 
         spec = self._worker_group.spec
         role = spec.role
 
-        log.info(f"[{role}] starting workers for function: {spec.fn.__name__}")
+        log.info(
+            f"[{role}] starting workers for entrypoint: {spec.get_entrypoint_name()}"
+        )
 
         self._initialize_workers(self._worker_group)
         monitor_interval = spec.monitor_interval
@@ -743,8 +806,8 @@ class SimpleElasticAgent(ElasticAgent):
         while True:
             assert self._worker_group.state != WorkerState.INIT
             time.sleep(monitor_interval)
-            monitor_result = self._monitor_workers(self._worker_group)
-            state = monitor_result.state
+            run_result = self._monitor_workers(self._worker_group)
+            state = run_result.state
             self._worker_group.state = state
 
             put_metric(f"workers.{role}.remaining_restarts", self._remaining_restarts)
@@ -756,7 +819,7 @@ class SimpleElasticAgent(ElasticAgent):
                     f" Waiting {self._exit_barrier_timeout} seconds for other agents to finish."
                 )
                 self._exit_barrier()
-                return monitor_result.ret_vals
+                return run_result
             elif state in {WorkerState.UNHEALTHY, WorkerState.FAILED}:
                 if self._remaining_restarts > 0:
                     log.info(
@@ -770,10 +833,7 @@ class SimpleElasticAgent(ElasticAgent):
                     self._stop_workers(self._worker_group)
                     self._worker_group.state = WorkerState.FAILED
                     self._exit_barrier()
-                    raise WorkerGroupFailureException(
-                        f"[{role}] exceeded max_restarts={spec.max_restarts}",
-                        monitor_result.exceptions,
-                    )
+                    return run_result
             elif state == WorkerState.HEALTHY:
                 # membership changes do not count as retries
                 num_nodes_waiting = rdzv_handler.num_nodes_waiting()
