@@ -219,20 +219,21 @@ job is invoking this launcher.
 """
 import logging
 import os
-import signal
-import subprocess
 import sys
 import uuid
 from argparse import REMAINDER, ArgumentParser
 
 import torch
 import torchelastic.rendezvous.registry as rdzv_registry
-from torchelastic import metrics
-from torchelastic.agent.server.api import WorkerGroupFailureException, WorkerSpec
+from torchelastic import metrics, events
+from torchelastic.agent.server.api import WorkerSpec, WorkerState
 from torchelastic.agent.server.local_elastic_agent import LocalElasticAgent
 from torchelastic.distributed.argparse_util import check_env, env
+from torchelastic.multiprocessing import Std
+from torchelastic.multiprocessing.errors import ChildFailedError, record
 from torchelastic.rendezvous import RendezvousParameters
 from torchelastic.rendezvous.etcd_server import EtcdServer
+from torchelastic.rendezvous.utils import _parse_rendezvous_config
 from torchelastic.utils.logging import get_logger
 
 
@@ -345,6 +346,36 @@ def parse_args(args):
         "it directly. Useful when the script is not a Python script.",
     )
 
+    parser.add_argument(
+        "--log_dir",
+        action=env,
+        type=str,
+        default=None,
+        help="base dir to use for log files (e.g. /var/log/torchelastic)"
+        " can reuse the same dir for multiple runs "
+        "(a unique job-level subdir is created with rdzv_id as the prefix)",
+    )
+
+    parser.add_argument(
+        "-r",
+        "--redirects",
+        action=env,
+        type=str,
+        default="0",
+        help="std streams to redirect into a log file in the log_dir"
+        " (e.g. [-r 3] redirects both stdout+stderr for all workers,"
+        " [-r 0:1,1:2] redirects stdout for local rank 0 and stderr for local rank 1)",
+    )
+
+    parser.add_argument(
+        "-t",
+        "--tee",
+        action=env,
+        type=str,
+        default="0",
+        help="tee std streams into a log file and also to console (see --redirects for format)",
+    )
+
     # positional
     parser.add_argument(
         "training_script",
@@ -374,50 +405,6 @@ def parse_min_max_nnodes(nnodes: str):
     return min_nodes, max_nodes
 
 
-def wrapper_fn(omp_num_threads, cmd):
-    # TODO get rid of this wrapper_fn when we switch over to torchelastic.multiprocessing
-    # the agent uses multiprocessing.spawn to create nproc_per_node
-    # instances of fn, and hence expects fn to be a callable
-    # since this launcher deals with user python scripts and executables
-    # we wrap the script/executable with this function which Popens
-    # the wrapped script/executable. This implies that for each
-    # worker we create two processes (wrapper_fn and fn).
-    # the process tree looks like the following:
-    #
-    # [launcher/agent]
-    #               |-- [wrapper_fn_0]
-    #               |               |-- [fn_0]
-    #               |-- [wrapper_fn_1]
-    #               |               |-- [fn_1]
-    #               |      ...
-    #               |      ...
-    #               |-- [wrapper_fn_k]
-    #               |               |-- [fn_k]
-    #
-
-    # set PyTorch distributed related environmental variables
-    if omp_num_threads is not None:
-        os.environ["OMP_NUM_THREADS"] = str(omp_num_threads)
-
-    process = subprocess.Popen(cmd)
-
-    # since we wrap the script process with this function (which runs as a
-    # subprocess of the agent) when the agent terminates this function
-    # due to some exception or membership change event we want the script
-    # to also get killed. If we do not register this exit handler
-    # the script process will get re-parented to the parent of this function
-    # (agent process) and we will end up with multiple copies of the script
-    # this should all go away with D20613415
-    def kill_script_pid(signum, frame):
-        process.terminate()
-
-    signal.signal(signal.SIGTERM, kill_script_pid)
-
-    process.wait()
-    if process.returncode != 0:
-        raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd)
-
-
 def determine_local_world_size(nproc_per_node: str):
     try:
         logging.info(f"Using nproc_per_node={nproc_per_node}.")
@@ -449,14 +436,27 @@ def determine_local_world_size(nproc_per_node: str):
         return num_proc
 
 
+def _construct_event(args) -> events.Event:
+    metadata = {
+        "rdzv_backend": args.rdzv_backend,
+        "run_id": args.run_id,
+        "role": args.role,
+    }
+    return events.Event(
+        name="torchelastic.main", source=events.EventSource.AGENT, metadata=metadata
+    )
+
+
+@record
 def main(args=None):
     # If ``args`` not passed, defaults to ``sys.argv[:1]``
     args = parse_args(args)
-
     min_nodes, max_nodes = parse_min_max_nnodes(args.nnodes)
     assert 0 < min_nodes <= max_nodes
     assert args.max_restarts >= 0
 
+    elastic_agent = None
+
     if args.standalone:
         etcd_server = EtcdServer()
         etcd_server.start()
@@ -473,7 +473,6 @@ def main(args=None):
         )
 
     nproc_per_node = determine_local_world_size(args.nproc_per_node)
-    omp_num_threads = None
     if "OMP_NUM_THREADS" not in os.environ and nproc_per_node > 1:
         omp_num_threads = 1
         print(
@@ -484,6 +483,8 @@ def main(args=None):
             f"your application as needed. \n"
             f"*****************************************"
         )
+        # This env variable will be passed down to the subprocesses
+        os.environ["OMP_NUM_THREADS"] = str(omp_num_threads)
 
     with_python = not args.no_python
     cmd = []
@@ -507,31 +508,44 @@ def main(args=None):
         run_id=args.rdzv_id,
         min_nodes=min_nodes,
         max_nodes=max_nodes,
-        **_parse_rdzv_conf(args.rdzv_conf),
+        **_parse_rendezvous_config(args.rdzv_conf),
     )
 
     rdzv_handler = rdzv_registry.get_rendezvous_handler(rdzv_parameters)
-
     try:
         spec = WorkerSpec(
             role=args.role,
             local_world_size=nproc_per_node,
-            fn=wrapper_fn,
-            args=(omp_num_threads, cmd),
+            entrypoint=cmd[0],
+            args=(*cmd[1:],),
             rdzv_handler=rdzv_handler,
             max_restarts=args.max_restarts,
             monitor_interval=args.monitor_interval,
+            redirects=Std.from_str(args.redirects),
+            tee=Std.from_str(args.tee),
         )
         metrics.initialize_metrics()
-        elastic_agent = LocalElasticAgent(spec, start_method=args.start_method)
-        elastic_agent.run(spec.role)
-    except WorkerGroupFailureException as e:
-        # user-code error; propagate the one from the smallest rank on this node
-        excs = e.get_worker_exceptions()
-        min_rank = min(excs.keys())
-        raise excs[min_rank] from e
+        elastic_agent = LocalElasticAgent(
+            spec=spec, start_method=args.start_method, log_dir=args.log_dir
+        )
+        run_result = elastic_agent.run(spec.role)
+        events.record(elastic_agent.get_agent_status_event(WorkerState.SUCCEEDED))
+        if run_result.is_failed():
+            # ChildFailedError is treated specially by @record
+            # if the error files for the failed children exist
+            # @record will copy the first error (root cause)
+            # to the error file of the launcher process
+            raise ChildFailedError(
+                name=args.training_script,
+                failures=run_result.failures,
+            )
+    except ChildFailedError:
+        raise
     except Exception:
-        # this is an agent (platform) error
+        if elastic_agent:
+            events.record(elastic_agent.get_agent_status_event(WorkerState.FAILED))
+        else:
+            events.record(_construct_event(args))
         raise
     finally:
         rdzv_handler.shutdown()
@@ -539,16 +553,6 @@ def main(args=None):
             etcd_server.stop()
 
 
-def _parse_rdzv_conf(conf_str: str):
-    rdzv_conf = {}
-    confs = conf_str.split(",")
-    for kvs in confs:
-        key = kvs.split("=")[0]
-        value = kvs.split("=")[1:]
-        rdzv_conf[key] = value
-    return rdzv_conf
-
-
 if __name__ == "__main__":
     logging.basicConfig(
         level=logging.INFO, format="[%(levelname)s] %(asctime)s %(module)s: %(message)s"
