@@ -13,6 +13,7 @@ import socket
 import subprocess
 import tempfile
 import time
+from typing import Optional
 
 import etcd
 from torchelastic.utils.logging import get_logger
@@ -59,14 +60,15 @@ def find_free_port():
     raise RuntimeError("Failed to create a socket")
 
 
-def stop_etcd(subprocess, data_dir):
+def stop_etcd(subprocess, data_dir: Optional[str] = None):
     if subprocess and subprocess.poll() is None:
         log.info(f"stopping etcd server")
         subprocess.terminate()
         subprocess.wait()
 
-    log.info(f"deleting etcd data dir: {data_dir}")
-    shutil.rmtree(data_dir, ignore_errors=True)
+    if data_dir:
+        log.info(f"deleting etcd data dir: {data_dir}")
+        shutil.rmtree(data_dir, ignore_errors=True)
 
 
 class EtcdServer:
@@ -101,7 +103,7 @@ class EtcdServer:
         etcd_binary_path: path of etcd server binary (see above for fallback path)
     """
 
-    def __init__(self):
+    def __init__(self, data_dir: Optional[str] = None):
         self._port = -1
         self._host = "localhost"
 
@@ -113,7 +115,9 @@ class EtcdServer:
         if not os.path.isfile(self._etcd_binary_path):
             self._etcd_binary_path = "etcd"
 
-        self._data_dir = tempfile.mkdtemp(prefix="torchelastic_etcd_data")
+        self._base_data_dir = (
+            data_dir if data_dir else tempfile.mkdtemp(prefix="torchelastic_etcd_data")
+        )
         self._etcd_cmd = None
         self._etcd_proc = None
 
@@ -138,7 +142,7 @@ class EtcdServer:
         """
         return f"{self._host}:{self._port}"
 
-    def start(self, timeout: int = 60) -> None:
+    def start(self, timeout: int = 60, num_retries: int = 3) -> None:
         """
         Starts the server, and waits for it to be ready. When this function
         returns the sever is ready to take requests.
@@ -146,11 +150,30 @@ class EtcdServer:
         Args:
             timeout: time (in seconds) to wait for the server to be ready
                 before giving up.
+            num_retries: number of retries to start the server. Each retry
+                will wait for max ``timeout`` before considering it as failed.
 
         Raises:
             TimeoutError: if the server is not ready within the specified timeout
         """
+        curr_retries = 0
+        while True:
+            try:
+                data_dir = os.path.join(self._base_data_dir, str(curr_retries))
+                os.makedirs(data_dir, exist_ok=True)
+                return self._start(data_dir, timeout)
+            except Exception as e:
+                curr_retries += 1
+                stop_etcd(self._etcd_proc)
+                log.warning(
+                    f"Failed to start etcd server, got error: {str(e)}, retrying"
+                )
+                if curr_retries >= num_retries:
+                    shutil.rmtree(self._base_data_dir, ignore_errors=True)
+                    raise
+        atexit.register(stop_etcd, self._etcd_proc, self._base_data_dir)
 
+    def _start(self, data_dir: str, timeout: int = 60) -> None:
         sock = find_free_port()
         sock_peer = find_free_port()
         self._port = sock.getsockname()[1]
@@ -162,7 +185,7 @@ class EtcdServer:
                     self._etcd_binary_path,
                     "--enable-v2",
                     "--data-dir",
-                    self._data_dir,
+                    data_dir,
                     "--listen-client-urls",
                     f"http://{self._host}:{self._port}",
                     "--advertise-client-urls",
@@ -178,7 +201,6 @@ class EtcdServer:
         sock.close()
         sock_peer.close()
         self._etcd_proc = subprocess.Popen(etcd_cmd, close_fds=True)
-        atexit.register(stop_etcd, self._etcd_proc, self._data_dir)
         self._wait_for_ready(timeout)
 
     def get_client(self) -> etcd.Client:
@@ -191,13 +213,18 @@ class EtcdServer:
             host=self._host, port=self._port, version_prefix="/v2", read_timeout=10
         )
 
-    def _wait_for_ready(self, timeout: int = 60):
+    def _wait_for_ready(self, timeout: int = 60) -> None:
         client = etcd.Client(
             host=f"{self._host}", port=self._port, version_prefix="/v2", read_timeout=5
         )
         max_time = time.time() + timeout
 
         while time.time() < max_time:
+            if self._etcd_proc.poll() is not None:
+                # etcd server process finished
+                raise RuntimeError(
+                    f"Etcd server process exited with the code: {self._etcd_proc.returncode}"
+                )
             try:
                 log.info(f"etcd server ready. version: {client.version}")
                 return
@@ -209,4 +236,5 @@ class EtcdServer:
         """
         Stops the server and cleans up auto generated resources (e.g. data dir)
         """
-        stop_etcd(self._etcd_proc, self._data_dir)
+        log.info("EtcdServer stop method called")
+        stop_etcd(self._etcd_proc, self._base_data_dir)
