@@ -14,7 +14,7 @@ def get_env_variable_or_raise(env_name: str) -> str:
     Tries to retrieve environment variable. Raises ``ValueError``
     if no environment variable found.
 
-    Arguments:
+    Args:
         env_name (str): Name of the env variable
     """
     value = os.environ.get(env_name, None)
